## DB Migrate
## ==========
##
## Simple tool to manage database migrations.

import std/[algorithm, json, logging, os, sequtils, sets, strutils, tables,
            times]
import db_connector/db_postgres
import docopt

type
  DbMigrateConfig* = object
    driver, connectionString: string
    sqlDirs: seq[string]
    logLevel: Level

  MigrationEntry* = tuple[name, upPath, downPath: string]

proc ensureMigrationsTableExists(conn: DbConn): void =
  let tableCount = conn.getValue(sql"""
SELECT COUNT(*) FROM information_schema.tables
WHERE table_name = 'migrations';""")

  if tableCount.strip == "0":
    info "Creating the migrations table as it does not already exist."
    conn.exec(sql("""
CREATE TABLE IF NOT EXISTS migrations (
  id SERIAL PRIMARY KEY,
  name VARCHAR NOT NULL,
  run_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW())"""))

proc rollbackWithErr (pgConn: DbConn, errMsg: string): void =
  pgConn.exec(sql"ROLLBACK")
  dbError(errMsg)

proc loadConfig*(filename: string): DbMigrateConfig =
  ## Load DbMigrateConfig from a file.
  let cfg = json.parseFile(filename)

  var logLevel: Level = lvlInfo
  if existsEnv("DBM_LOG_LEVEL"):
    let idx = find(LevelNames, $getEnv("DBM_LOG_LEVEL").toUpper)
    logLevel = if idx == -1: lvlInfo else: (Level)(idx)
  elif cfg.hasKey("logLevel"):
    let idx = find(LevelNames, cfg["logLevel"].getStr.toUpper)
    logLevel = if idx == -1: lvlInfo else: (Level)(idx)

  return DbMigrateConfig(
    driver:
      if existsEnv("DATABASE_DRIVER"): $getEnv("DATABASE_DRIVER")
      elif cfg.hasKey("driver"): cfg["driver"].getStr
      else: "postres",
    connectionString:
      if existsEnv("DATABASE_URL"): $getEnv("DATABASE_URL")
      elif cfg.hasKey("connectionString"): cfg["connectionString"].getStr
      else: "",
    sqlDirs:
      if existsEnv("MIGRATIONS_DIRS"): getEnv("MIGRATIONS_DIRS").split(';')
      elif cfg.hasKey("sqlDirs"): cfg["sqlDirs"].getElems.mapIt(it.getStr)
      else: @["migrations"],
    logLevel: logLevel)

proc createMigration*(config: DbMigrateConfig, migrationName: string): MigrationEntry =
  ## Create a new set of database migration files.
  let timestamp = now().format("yyyyMMddHHmmss")
  let filenamePrefix =  timestamp & "-" & migrationName

  let migration = (
    name: filenamePrefix & "-up.sql",
    upPath: joinPath(config.sqlDirs[0], filenamePrefix & "-up.sql"),
    downPath: joinPath(config.sqlDirs[0], filenamePrefix & "-down.sql"))

  let scriptDesc = migrationName & " (" & timestamp & ")"

  let upFile = open(migration.upPath, fmWrite)
  let downFile = open(migration.downPath, fmWrite)

  upFile.writeLine "-- UP script for " & scriptDesc
  downFile.writeLine "-- DOWN script for " & scriptDesc

  upFile.close()
  downFile.close()

  return migration

proc diffMigrations*(
    pgConn: DbConn,
    config: DbMigrateConfig
  ): tuple[
    available: TableRef[string, MigrationEntry],
    run: seq[string],
    notRun, missing: seq[MigrationEntry] ] =

  debug "diffMigrations: inspecting database and configured directories " &
    "for migrations"

  # Query the database to find out what migrations have been run.
  var migrationsRun = initHashSet[string]()

  for row in pgConn.fastRows(sql"SELECT * FROM migrations ORDER BY name", @[]):
    migrationsRun.incl(row[1])

  # Inspect the filesystem to see what migrations are available.
  var migrationsAvailable = newTable[string, MigrationEntry]()
  for sqlDir in config.sqlDirs:
    debug "Looking in " & sqlDir
    for filePath in walkFiles(joinPath(sqlDir, "*.sql")):
      debug "Saw migration file: " & filePath
      var migrationName = filePath.extractFilename
      migrationName.removeSuffix("-up.sql")
      migrationName.removeSuffix("-down.sql")
      migrationsAvailable[migrationName] = (
        name: migrationName,
        upPath: joinPath(sqlDir, migrationName) & "-up.sql",
        downPath: joinPath(sqlDir, migrationName) & "-down.sql")

  # Diff with the list of migrations that we have in our migrations
  # directory.
  let migrationsInOrder =
    toSeq(migrationsAvailable.keys).sorted(system.cmp)

  var migrationsNotRun = newSeq[MigrationEntry]()
  var missingMigrations = newSeq[MigrationEntry]()

  for migName in migrationsInOrder:
    if not migrationsRun.contains(migName):
      migrationsNotRun.add(migrationsAvailable[migName])

    # if we've already seen some migrations that have not been run, but this
    # one has been, that means we have a gap and are missing migrations
    elif migrationsNotRun.len > 0:
      missingMigrations.add(migrationsNotRun)
      migrationsNotRun = newSeq[MigrationEntry]()

  result = (available: migrationsAvailable,
          run: toSeq(migrationsRun.items).sorted(system.cmp),
          notRun: migrationsNotRun,
          missing: missingMigrations)

  debug "diffMigration: Results" &
    "\n\tavailable: " & $toSeq(result[0].keys) &
    "\n\trun:       " & $result[1] &
    "\n\tnotRun:    " & $(result[2].mapIt(it.name)) &
    "\n\tmissing:   " & $(result[3].mapIt(it.name))


proc readStatements*(filename: string): seq[SqlQuery] =
  result = @[]
  var stmt: string = ""

  for line in filename.lines:
    let l = line.strip
    if l.len == 0 or l.startsWith("--"): continue

    let parts = line.split(';')
    stmt &= "\n" & parts[0]

    if parts.len > 1:
      result.add(sql(stmt & ";"))
      stmt = parts[1] & "";

  if stmt.strip.len > 0: result.add(sql(stmt))

proc up*(
    pgConn: DbConn,
    config: DbMigrateConfig,
    toRun: seq[MigrationEntry]): seq[MigrationEntry] =
  var migrationsRun = newSeq[MigrationEntry]()
  # Begin a transaction.
  pgConn.exec(sql"BEGIN")

  # Apply each of the migrations.
  for migration in toRun:
    info migration.name

    if not migration.upPath.fileExists:
      pgConn.rollbackWithErr "Can not find UP file for " & migration.name &
        ". Expected '" & migration.upPath & "'."

    let statements = migration.upPath.readStatements

    try:
      for statement in statements:
        pgConn.exec(statement)
      pgConn.exec(sql"INSERT INTO migrations (name) VALUES (?);", migration.name)
    except DbError:
      pgConn.rollbackWithErr "Migration '" & migration.name & "' failed:\n\t" &
        getCurrentExceptionMsg()

    migrationsRun.add(migration)

  pgConn.exec(sql"COMMIT")

  return migrationsRun

proc down*(
    pgConn: DbConn,
    config: DbMigrateConfig,
    migrationsToDown: seq[MigrationEntry]): seq[MigrationEntry] =
  var migrationsDowned = newSeq[MigrationEntry]()

  pgConn.exec(sql"BEGIN")

  for migration in migrationsToDown:
    info migration.name

    if not migration.downPath.fileExists:
      pgConn.rollbackWithErr "Can not find DOWN file for " & migration.name &
        ". Expected '" & migration.downPath & "'."

    let statements = migration.downPath.readStatements

    try:
      for statement in statements: pgConn.exec(statement)
      pgConn.exec(sql"DELETE FROM migrations WHERE name = ?;", migration.name)
    except DbError:
      pgConn.rollbackWithErr "Migration '" & migration.name & "' failed:\n\t" &
        getCurrentExceptionMsg()

    migrationsDowned.add(migration)

  pgConn.exec(sql"COMMIT")
  return migrationsDowned

when isMainModule:
  let doc = """
Usage:
  db_migrate [options] create <migration-name>
  db_migrate [options] up [<count>]
  db_migrate [options] down [<count>]
  db_migrate [options] init <schema-name>
  db_migrate (-V | --version)
  db_migrate (-h | --help)

Options:

  -c --config <config-file>   Use the given configuration file (defaults to
                              "database.json").

  -h --help                   Show this usage information.

  -q --quiet                  Suppress log information.

  -v --verbose                Print detailed log information.

     --very-verbose           Print very detailed log information.

  -V --version                Print the tools version information.
"""

  # Parse arguments
  let args = docopt(doc, version = "db-migrate (Nim) 0.3.2\nhttps://git.jdb-software.com/jdb/db-migrate")

  let exitErr = proc(msg: string): void =
    fatal("db_migrate: " & msg)
    quit(QuitFailure)

  # Load configuration file
  let configFilename =
    if args["--config"]: $args["--config"]
    else: "database.json"

  var config: DbMigrateConfig
  try:
    config = loadConfig(configFilename)
  except IOError:
    writeLine(stderr, "db_migrate: Cannot open config file: " & configFilename)
  except:
    writeLine(stderr, "db_migrate: Error parsing config file: " &
      configFilename & "\L\t" & getCurrentExceptionMsg())


  logging.addHandler(newConsoleLogger())
  if args["--quiet"]: logging.setLogFilter(lvlError)
  elif args["--very-verbose"]: logging.setLogFilter(lvlAll)
  elif args["--verbose"]: logging.setlogFilter(lvlDebug)
  else: logging.setLogFilter(config.logLevel)

  # Check for migrations directory
  for sqlDir in config.sqlDirs:
    if not dirExists sqlDir:
      try:
        warn "SQL directory '" & sqlDir &
          "' does not exist and will be created."
        createDir sqlDir
      except IOError:
        exitErr "Unable to create directory: " & sqlDir & ":\L\T" & getCurrentExceptionMsg()

  # Execute commands
  if args["create"]:
    try:
      let newMigration = createMigration(config, $args["<migration-name>"])
      info "Created new migration files:"
      info "\t" & newMigration.upPath
      info "\t" & newMigration.downPath
    except IOError:
      exitErr "Unable to create migration scripts: " & getCurrentExceptionMsg()

  else:
    let pgConn: DbConn =
      try: open("", "", "", config.connectionString)
      except DbError:
        exitErr "Unable to connect to the database: " & getCurrentExceptionMsg()
        (DbConn)(nil)

    pgConn.ensureMigrationsTableExists

    let (available, run, notRun, missing) = diffMigrations(pgConn, config)

    # Make sure we have no gaps (database is in an unknown state)
    if missing.len > 0:
      exitErr "Database is in an inconsistent state. Migrations have been " &
        "run that are not sequential."

    if args["up"]:
      try:
        let count = if args["<count>"]: parseInt($args["<count>"]) else: high(int)
        let toRun = if count < notRun.len: notRun[0..<count] else: notRun
        let migrationsRun = pgConn.up(config, toRun)
        if count < high(int): info "Went up " & $(migrationsRun.len) & "."
      except DbError:
        exitErr "Unable to migrate database: " & getCurrentExceptionMsg()

    elif args["down"]:
      try:
        let count = if args["<count>"]: parseInt($args["<count>"]) else: 1
        let toRunNames = if count < run.len: run.reversed[0..<count] else: run.reversed
        let toRun = toRunNames.mapIt(available[it])
        let migrationsRun = pgConn.down(config, toRun)
        info "Went down " & $(migrationsRun.len) & "."
      except DbError:
        exitErr "Unable to migrate database: " & getCurrentExceptionMsg()

    elif args["init"]: discard

    let newResults = diffMigrations(pgConn, config)
    if newResults.notRun.len > 0:
      info "Database is behind by " & $(newResults.notRun.len) & " migrations."
    else: info "Database is up to date."
